from aurem.pickers import REC, AIC
from obspy import read, UTCDateTime


def test_aurem_rec():
    """ Test REC function aganist obspy.read() Z channel
    """
    errors = []
    st = read()

    # Process
    st.filter('highpass', freq=2, corners=4)
    st.trim(st[0].stats.starttime + 1, UTCDateTime("2009-08-24T00:20:11"))

    # Create instance + pick
    recobj = REC(st, channel="*Z")
    recobj.work()

    # Extract
    pt = recobj.get_pick()
    idx = recobj.get_pick_index()
    # recobj.plot()

    # --- Test
    if not isinstance(pt, UTCDateTime):
        errors.append("Wrong out picktime object")
    if pt != UTCDateTime("2009-08-24T00:20:07.700000"):
        errors.append("PickTime do not match!")
    if idx != 370:
        errors.append("PickTime IDX do not match!")
    #
    assert not errors, "Errors occured:\n{}".format("\n".join(errors))


def test_aurem_aic():
    """ Test AIC function aganist obspy.read() Z channel
    """
    errors = []
    st = read()

    # Process
    st.filter('highpass', freq=2, corners=4)
    st.trim(st[0].stats.starttime + 1, UTCDateTime("2009-08-24T00:20:11"))

    # Create Instance + picks
    aicobj = AIC(st, channel="*Z")
    aicobj.work()

    # Extract
    pt = aicobj.get_pick()
    idx = aicobj.get_pick_index()
    # aicobj.plot()

    # --- Test
    if not isinstance(pt, UTCDateTime):
        errors.append("Wrong out picktime object")
    if pt != UTCDateTime("2009-08-24T00:20:07.700000"):
        errors.append("PickTime do not match!")
    if idx != 370:
        errors.append("PickTime IDX do not match!")
    #
    assert not errors, "Errors occured:\n{}".format("\n".join(errors))



import copy
import numpy as np
import logging
import matplotlib.pyplot as plt
plt.style.context('fivethirtyeight')


logger = logging.getLogger(__name__)


# =======================  Common

def _normalize_trace(workList, rangeVal=[-1, 1]):
    """Normalize array between 2 values

    This function normalize the trace between 2 numbers defined in input

    Inputs:
        workList (np.ndarray, list, tuple): input array to normalize
        rangeVal (list, tuple): min and max range of normalization

    Returns:
        worklist (np.ndarray, list, tuple): normalized input array.

    """
    minVal = min(workList)
    maxVal = max(workList)
    workList[:] = [((x - minVal) / (maxVal - minVal)) *
                   (rangeVal[1] - rangeVal[0]) for x in workList]
    workList = workList + rangeVal[0]
    return workList


# =======================  Main

def plot_rec(rec_obj,
             plot_ax=None,
             plot_cf=False,
             plot_pick=True,
             plot_additional_picks={},
             normalize=True,
             axtitle="AUREM picker: REC",
             show=False):
    """Plotting function for REC auto-regressive method.

    Args:
        rec_obj (aurem.picker.REC): REC picker object

    Optionals:
        plot_ax (matplotlib.pyplot.axes): handle of the axes where plot
        plot_cf (bool): either plot or not the CF of REC
        plot_pick (bool): either plot or not the final REC-pick
        plot_additional_picks (dict): key-args dict with
            'name': UTCDateTime scheme.
        normalize (bool): if True normalize between -1,1 the trace and
            from 0 to 1 the (optional) REC-CF.
        axtitle (str): axis title
        show (bool): if True plot on screen the resulting figure

    Returns:
        inax (matplotlib.pyplot.axes): handle of the active axis

    """
    if not plot_ax:
        inax = plt.axes()
    else:
        inax = plot_ax

    # Creating time vector and trace data
    newTrace = rec_obj.wt.copy()
    tv = newTrace.times()
    td = newTrace.data

    if normalize:
        td = _normalize_trace(td, rangeVal=[-1, 1])

    # -------------------------- Carachteristic function
    if plot_cf:
        if rec_obj.recfn.any():
            _wa = copy.deepcopy(rec_obj.recfn)
            # GoingToC: replace INF at the start and end with adiacent
            #           values for plotting reasons
            _wa[0] = _wa[1]
            _wa[-1] = _wa[-2]

            if normalize:
                _tmp = _normalize_trace(_wa, rangeVal=[0, 1])
            #
            zeropad = len(td) - len(_tmp)
            #
            inax.plot(tv, np.pad(_tmp, (0, zeropad), mode='constant',
                                 constant_values=(np.nan,)),
                      color="teal",
                      linewidth=1,
                      linestyle=':',
                      label="CF")
        else:
            logger.warning("Missing REC CF, skipping ...")

    # -------------------------- Additional Picks
    my_color_list_add = ['lime',
                         'forestgreen',
                         'limegreen',
                         'darkgreen']
    col_idx = 0
    if plot_additional_picks and isinstance(plot_additional_picks, dict):
        for _kk, _pp in plot_additional_picks.items():
            inax.axvline(_pp - rec_obj.wt.stats.starttime,
                         color=my_color_list_add[col_idx],
                         linewidth=1.5,
                         label=_kk)
            col_idx += 1

    # -------------------------- Final Pick
    if plot_pick:
        if rec_obj.pick:
            inax.axvline(rec_obj.pick - rec_obj.wt.stats.starttime,
                         color="teal",
                         linestyle="-",
                         linewidth=2,
                         label="PICK")
        else:
            logger.warning("Missing REC PICK, skipping ...")

    # -------------------------- Plot trace
    inax.plot(tv, td, "k", label="trace")
    inax.set_xlabel("time (s)")
    inax.set_ylabel("counts")
    inax.legend(loc='lower left')
    inax.set_title(axtitle, {'fontsize': 16, 'fontweight': 'bold'})

    if show:
        plt.tight_layout()
        plt.show()
    #
    return inax


def plot_aic(aic_obj,
             plot_ax=None,
             plot_cf=False,
             plot_pick=True,
             plot_additional_picks={},
             normalize=True,
             axtitle="AUREM picker: AIC",
             show=False):
    """Plotting function for AIC auto-regressive method.

    Args:
        aic_obj (aurem.picker.AIC): AIC picker object

    Optionals:
        plot_ax (matplotlib.pyplot.axes): handle of the axes where plot
        plot_cf (bool): either plot or not the CF of AIC
        plot_pick (bool): either plot or not the final AIC-pick
        plot_additional_picks (dict): key-args dict with
            'name': UTCDateTime scheme.
        normalize (bool): if True normalize between -1,1 the trace and
            from 0 to 1 the (optional) AIC-CF.
        axtitle (str): axis title
        show (bool): if True plot on screen the resulting figure

    Returns:
        inax (matplotlib.pyplot.axes): handle of the active axis

    """

    if not plot_ax:
        inax = plt.axes()
    else:
        inax = plot_ax

    # Creating time vector and trace data
    newTrace = aic_obj.wt.copy()
    tv = newTrace.times()
    td = newTrace.data

    if normalize:
        td = _normalize_trace(td, rangeVal=[-1, 1])

    # -------------------------- Carachteristic function
    if plot_cf:
        if aic_obj.aicfn.any():
            _wa = copy.deepcopy(aic_obj.aicfn)
            # GoingToC: replace INF at the start and end with adiacent
            #           values for plotting reasons
            _wa[0] = _wa[1]
            _wa[-1] = _wa[-2]

            if normalize:
                _tmp = _normalize_trace(_wa, rangeVal=[0, 1])
            #
            zeropad = len(td) - len(_tmp)
            #
            inax.plot(tv, np.pad(_tmp, (0, zeropad), mode='constant',
                                 constant_values=(np.nan,)),
                      color="teal",
                      linewidth=1,
                      linestyle=':',
                      label="CF")
        else:
            logger.warning("Missing AIC CF, skipping ...")

    # -------------------------- Additional Picks
    my_color_list_add = ['lime',
                         'forestgreen',
                         'limegreen',
                         'darkgreen']
    col_idx = 0
    if plot_additional_picks and isinstance(plot_additional_picks, dict):
        for _kk, _pp in plot_additional_picks.items():
            inax.axvline(_pp - aic_obj.wt.stats.starttime,
                         color=my_color_list_add[col_idx],
                         linewidth=1.5,
                         label=_kk)
            col_idx += 1

    # -------------------------- Final Pick
    if plot_pick:
        if aic_obj.pick:
            inax.axvline(aic_obj.pick - aic_obj.wt.stats.starttime,
                         color="teal",
                         linestyle="-",
                         linewidth=2,
                         label="PICK")
        else:
            logger.warning("Missing AIC PICK, skipping ...")

    # -------------------------- Plot trace
    inax.plot(tv, td, "k", label="trace")
    inax.set_xlabel("time (s)")
    inax.set_ylabel("counts")
    inax.legend(loc='lower left')
    inax.set_title(axtitle, {'fontsize': 16, 'fontweight': 'bold'})

    if show:
        plt.tight_layout()
        plt.show()
    #
    return inax


# ================== TIPS
# To create a continuos discrete list of colors with matplotlib

# import matplotlib.cm as mplcm
# import matplotlib.colors as colors
#    NUM_COLORS = len(pickTime_UTC.keys()) - 2  # MB: - mean and  -median
#    cm = plt.get_cmap('gist_rainbow')
#    cNorm = colors.Normalize(vmin=0, vmax=NUM_COLORS-1)
#    scalarMap = mplcm.ScalarMappable(norm=cNorm, cmap=cm)
#    inax.set_prop_cycle(color=[scalarMap.to_rgba(i)
#                               for i in range(NUM_COLORS)])



import logging
import numpy as np
import pathlib
import ctypes as C
import copy
from obspy import UTCDateTime
#
from aurem import plotting as AUPL

logger = logging.getLogger(__name__)


# -------------------------------------------  Load and Setup C library
MODULEPATH = pathlib.Path(__file__).parent.absolute()

libname = tuple(MODULEPATH.glob("src/aurem_clib.*.so"))[0]
myclib = C.CDLL(libname)

# AIC
myclib.aicp.restype = C.c_int
myclib.aicp.argtypes = [np.ctypeslib.ndpointer(
                                        dtype=np.float32, ndim=1,
                                        flags='C_CONTIGUOUS'), C.c_int,
                        # OUT
                        np.ctypeslib.ndpointer(
                                        dtype=np.float32, ndim=1,
                                        flags='C_CONTIGUOUS'),
                        C.POINTER(C.c_int)]

# REC
myclib.recp.restype = C.c_int
myclib.recp.argtypes = [np.ctypeslib.ndpointer(
                                        dtype=np.float32, ndim=1,
                                        flags='C_CONTIGUOUS'), C.c_int,
                        # OUT
                        np.ctypeslib.ndpointer(
                                        dtype=np.float32, ndim=1,
                                        flags='C_CONTIGUOUS'),
                        C.POINTER(C.c_int)]
# ---------------------------------------------------------------------


class REC(object):
    """ Initialize the Reciprocal-Based picker class.

    The picker takes in inputs only the obspy.Stream objects.
    Being completely automatic, only the channel parameter to
    specify the working trace is accepted.

    For an exhaustive description of the picker's parameters the
    reader is referred to the reference paper.

    Args:
        stream (obspy.Stream): working Stream containing the working
            traces

    Optional:
        querykey (str): the following key-value parameters will be
            used to query the Stream at class initialization. Use the
            class method `set_working_trace` before running the work
            method.

    Note:
        If no query key-args given, the class will set the
        working trace as the 1st stream-trace !!!

    References:
        Ramin Madarshahian, Paul Ziehl, and Michael D. Todd (2020),
            Bayesian Estimation of Acoustic Emission Arrival Times for
            Source Localization.
            DOI: https://doi.org/10.1007/978-3-030-12075-7_13

    """
    def __init__(self, stream, **streamselect):
        self.wt = None  # just to define variable --> set_working_trace
        self.st = stream.copy()
        self.set_working_trace(**streamselect)
        #
        self.recfn = None
        self.idx = None
        self.pick = None

    def work(self):
        """ Core method ruling the picking workflow

        This method will create the CF and store pick index and UTC time

        Note:
            Now the CF's calcultation of CF and index extraction are
            contained in a faster C routine

        """
        if self.wt:
            pminidx = C.c_int()
            tmparr = np.ascontiguousarray(self.wt.data, np.float32)
            self.recfn = np.zeros(self.wt.data.size - 1,
                                  dtype=np.float32, order="C")
            ret = myclib.recp(tmparr, self.wt.data.size,
                              self.recfn, C.byref(pminidx))
            if ret != 0:
                raise MemoryError("Something wrong with REC picker C-routine")
            #
            self.idx = pminidx.value
            if self.idx != 0 and isinstance(self.idx, int):
                # pick found
                logger.debug("REC found pick")
                self.pick = (self.wt.stats.starttime +
                             self.wt.stats.delta * self.idx)
            else:
                # if idx == 0, it means inside C routine it didn't pick
                logger.debug("REC didn't found pick")
                self.pick = None
        else:
            raise AttributeError("Missing working trace. Use the class "
                                 "`set_working_trace` mothod first!")

    def set_working_trace(self, **streamselect):
        """ Select the trace from the class obspy.Stream

        The optional query parameters are the same supported by the
        obspy.Stream.selection() method

        Note:
            If no query key-args given, the class will set the
            working trace as the 1st stream-trace !!!

        """
        if streamselect and not isinstance(streamselect, dict):
            raise TypeError("Please specify stream select key-args query!")
        #
        self.wt = self.st.select(**streamselect)[0]

    def get_rec_function(self, mode="real"):
        """ Return the REC charachteristic function.

        Utility function to return aic function for either

        Optional:
            mode (str): either "real" or "plot". If "real", the original
                CF is returned. If plot option, the method return a
                deepcopy of the original CF. The "plot" method is
                indicated for custom plots.

        """
        if isinstance(self.recfn, np.ndarray) and self.recfn.size > 0:
            if mode.lower() == 'real':
                return self.recfn
            elif mode.lower() == 'plot':
                _wa = copy.deepcopy(self.recfn)
                # GoingToC: replace INF at the start and end with
                #           adiacent values for plotting reasons
                _wa[0] = _wa[1]
                _wa[-1] = _wa[-2]
                return _wa
        else:
            logger.warning("Missing CF FUNCTION! " +
                           "Run the work method first!")

    def get_pick_index(self):
        """ Return the pick-sample index

        Returns:
            idx (int): index of the pick

        """
        if self.idx:
            return self.idx
        else:
            logger.warning("Missing INDEX! " +
                           "Run the work method first!")

    def get_pick(self):
        """ Return the pick-UTC time

        Returns:
            pick (obspy.UTCDateTime): UTC time of the pick

        """
        if self.pick:
            if isinstance(self.pick, UTCDateTime):
                return self.pick
            else:
                raise TypeError
        else:
            logger.warning("Missing PICK! " +
                           "Run the work method first!")

    def plot(self):
        """ Wrapper around the plot_rec function.

        This method will quickly produce a plot with CF and picks
        For different type of plots, please use the plotting.plot_rec
        module instead.

        Returns:
            ax (matplotlib.pyplot.axes): axis object of the plot

        """
        ax = AUPL.plot_rec(self,
                           plot_ax=None,
                           plot_cf=True,
                           plot_pick=True,
                           plot_additional_picks={},
                           normalize=True,
                           axtitle="REC picks",
                           show=True)
        return ax


class AIC(object):
    """ Initialize the Akaike-Information-Criteria picker class.

    The picker takes in inputs only the obspy.Stream objects.
    Being completely automatic, only the channel parameter to
    specify the working trace is accepted.

    For an exhaustive description of the picker's parameters the
    reader is referred to the reference paper.

    Args:
        stream (obspy.Stream): working Stream containing the working
            traces

    Optional:
        querykey (str): the following key-value parameters will be
            used to query the Stream at class initialization. Use the
            class method `set_working_trace` before running the work
            method.

    Note:
        If no query key-args given, the class will set the
        working trace as the 1st stream-trace !!!

    References:
        Maeda, Naoki. 1985. “A Method for Reading and Checking Phase
            Time in Auto-Processing System of Seismic Wave Data.”
            Zisin (Journal of the Seismological Society of Japan.
            2nd Ser.) 38 (3): 365–79.
            DOI: https://doi.org/10.4294/zisin1948.38.3_365

    """
    def __init__(self, stream, **streamselect):
        self.wt = None  # just to define variable --> set_working_trace
        self.st = stream.copy()
        self.set_working_trace(**streamselect)
        #
        self.aicfn = None
        self.idx = None
        self.pick = None

    def work(self):
        """ Core method ruling the picking workflow

        This method will create the CF and store pick index and UTC time

        Note:
            Now the CF's calcultation of CF and index extraction are
            contained in a faster C routine

        """
        if self.wt:
            pminidx = C.c_int()
            tmparr = np.ascontiguousarray(self.wt.data, np.float32)
            self.aicfn = np.zeros(self.wt.data.size - 1,
                                  dtype=np.float32, order="C")
            ret = myclib.aicp(tmparr, self.wt.data.size,
                              self.aicfn, C.byref(pminidx))
            if ret != 0:
                raise MemoryError("Something wrong with AIC picker C-routine")
            #
            self.idx = pminidx.value
            if self.idx != 0 and isinstance(self.idx, int):
                # pick found
                logger.debug("AIC found pick")
                self.pick = (self.wt.stats.starttime +
                             self.wt.stats.delta * self.idx)
            else:
                # if idx == 0, it means inside C routine it didn't pick
                logger.debug("AIC didn't found pick")
                self.pick = None
        else:
            raise AttributeError("Missing working trace. Use the class "
                                 "`set_working_trace` mothod first!")

    def set_working_trace(self, **streamselect):
        """ Select the trace from the class obspy.Stream

        The optional query parameters are the same supported by the
        obspy.Stream.selection() method

        Note:
            After the selection-query stace of class stream. It will
            select the first trace of filtered stream

        """
        if streamselect and not isinstance(streamselect, dict):
            raise TypeError("Please specify stream select key-args query!")
        #
        self.wt = self.st.select(**streamselect)[0]

    def get_aic_function(self, mode="real"):
        """ Return the AIC charachteristic function.

        Utility function to return aic function for either

        Optional:
            mode (str): either "real" or "plot". If "real", the original
                CF is returned. If plot option, the method return a
                deepcopy of the original CF. The "plot" method is
                indicated for custom plots.

        Note:
            If no query key-args given, the class will set the
            working trace as the 1st stream-trace !!!

        """
        if isinstance(self.aicfn, np.ndarray) and self.aicfn.size > 0:
            if mode.lower() == 'real':
                return self.aicfn
            elif mode.lower() == 'plot':
                _wa = copy.deepcopy(self.aicfn)
                # GoingToC: replace INF at the start and end with
                #           adiacent values for plotting reasons
                _wa[0] = _wa[1]
                _wa[-1] = _wa[-2]
                return _wa
            raise ValueError("Mode must be either 'real' or 'plot'")
        else:
            raise AttributeError("Missing CF FUNCTION! " +
                                 "Run the work method first!")

    def get_pick_index(self):
        """ Return the pick-sample index

        Returns:
            idx (int): index of the pick

        """
        if self.idx is None:
            raise AttributeError("Missing INDEX! " +
                                 "Run the work method first!")
        #
        if isinstance(self.idx, int):
            return self.idx
        else:
            raise TypeError("INDEX is not an integer: %s" %
                            str(type(self.idx)))

    def get_pick(self):
        """ Return the pick-UTC time

        Returns:
            pick (obspy.UTCDateTime): UTC time of the pick

        """
        if self.idx is None:
            raise AttributeError("Missing PICK! " +
                                 "Run the work method first!")
        else:
            # At this stage, it can be already UTCDateTime or None.
            return self.pick

    def plot(self):
        """ Wrapper around the plot_aic function.

        This method will quickly produce a plot with CF and picks
        For different type of plots, please use the plotting.plot_rec
        module instead.

        Returns:
            ax (matplotlib.pyplot.axes): axis object of the plot

        """
        ax = AUPL.plot_aic(self,
                           plot_ax=None,
                           plot_cf=True,
                           plot_pick=True,
                           plot_additional_picks={},
                           normalize=True,
                           axtitle="AIC picks",
                           show=True)
        return ax



from setuptools import setup, find_packages, Extension

with open("README.md", "r") as fh:
    long_description = fh.read()

with open("requirements.txt") as f:
    required_list = f.read().splitlines()


cmodule = Extension('aurem/src/aurem_clib',
                    sources=['aurem/src/aurem_clib.c'],
                    extra_compile_args=["-O3"])

setup(
    name="aurem",
    version="1.1.0",
    author="Matteo Bagagli",
    author_email="matteo.bagagli@erdw.ethz.com",
    description="AUto REgressive Models for seismic phase picking",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/mbagagli/aurem",
    python_requires='>=3.6',
    install_requires=required_list,
    packages=find_packages(),
    package_data={"aurem": ['src/*.c']},
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: Unix",
        "Operating System :: MacOS",
        "Intended Audience :: Science/Research",
    ],
    ext_modules=[cmodule]
)


# ========== HELP

# In this example, setup() is called with additional meta-information,
# which is recommended when distribution packages have to be built.
# For the extension itself, it specifies preprocessor defines,
# include directories, library directories, and libraries.
# Depending on the compiler, distutils passes this information in
# different ways to the compiler.
# For example, on Unix, this may result in the compilation commands

# ```
# gcc -DNDEBUG -g -O3 -Wall -Wstrict-prototypes -fPIC -DMAJOR_VERSION=1 \
#     -DMINOR_VERSION=0 -I/usr/local/include \
#     -I/usr/local/include/python2.2 -c demo.c \
#     -o build/temp.linux-i686-2.2/demo.o

# gcc -shared build/temp.linux-i686-2.2/demo.o -L/usr/local/lib -ltcl83
#     -o build/lib.linux-i686-2.2/demo.so
# ```

# These lines are for demonstration purposes only; distutils users
# should trust that distutils gets the invocations right.

# The manifest template has one command per line, where each command specifies a set of files to include or exclude from the source distribution. For an example, again we turn to the Distutils’ own manifest template:

# include *.txt
# recursive-include examples *.txt *.py
# prune examples/sample?/build

# The meanings should be fairly clear: include all files in the
# distribution root matching *.txt, all files anywhere under the examples
# directory matching *.txt or *.py, and exclude all directories matching
# examples/sample?/build. All of this is done after the standard include
# set, so you can exclude files from the standard set with explicit
# instructions in the manifest template. (Or, you can use the --no-defaults
# option to disable the standard set entirely.)
# There are several other commands available in the manifest template
# mini-language; see section Creating a source distribution:
# the sdist command.
